@@ -5,6 +5,7 @@ use std::{
     cmp::min,
     fmt,
     hash::{Hash, Hasher},
+    num::NonZeroU32,
 };
 use strsim::normalized_levenshtein;
 
@@ -44,8 +45,9 @@ pub struct Vertex<'a> {
     pub lhs_syntax: Option<&'a Syntax<'a>>,
     pub rhs_syntax: Option<&'a Syntax<'a>>,
     parents: Stack<EnteredDelimiter<'a>>,
-    lhs_parent_id: Option<u32>,
-    rhs_parent_id: Option<u32>,
+    lhs_parent_id: Option<NonZeroU32>,
+    rhs_parent_id: Option<NonZeroU32>,
+    can_pop_either: bool,
 }
 
 impl<'a> PartialEq for Vertex<'a> {
@@ -70,6 +72,12 @@ impl<'a> PartialEq for Vertex<'a> {
             // the graph size relative to tree depth.
             && self.lhs_parent_id == other.lhs_parent_id
             && self.rhs_parent_id == other.rhs_parent_id
+            // We do want to distinguish whether we can pop each side
+            // independently though. Without this, if we find a case
+            // where we can pop sides together, we don't consider the
+            // case where we get a better diff by popping each side
+            // separately.
+            && self.can_pop_either == other.can_pop_either
     }
 }
 impl<'a> Eq for Vertex<'a> {}
@@ -81,6 +89,7 @@ impl<'a> Hash for Vertex<'a> {
 
         self.lhs_parent_id.hash(state);
         self.rhs_parent_id.hash(state);
+        self.can_pop_either.hash(state);
     }
 }
 
@@ -106,7 +115,11 @@ impl<'a> fmt::Debug for EnteredDelimiter<'a> {
     fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
         let desc = match self {
             EnteredDelimiter::PopEither((lhs_delims, rhs_delims)) => {
-                format!("PopEither({}, {})", lhs_delims.size(), rhs_delims.size())
+                format!(
+                    "PopEither(lhs count: {}, rhs count: {})",
+                    lhs_delims.size(),
+                    rhs_delims.size()
+                )
             }
             EnteredDelimiter::PopBoth(_) => "PopBoth".to_string(),
         };
@@ -122,6 +135,13 @@ fn push_both_delimiters<'a>(
     entered.push(EnteredDelimiter::PopBoth((lhs_delim, rhs_delim)))
 }
 
+fn can_pop_either_parent(entered: &Stack<EnteredDelimiter>) -> bool {
+    match entered.peek() {
+        Some(EnteredDelimiter::PopEither(_)) => true,
+        _ => false,
+    }
+}
+
 fn try_pop_both<'a>(
     entered: &Stack<EnteredDelimiter<'a>>,
 ) -> Option<(&'a Syntax<'a>, &'a Syntax<'a>, Stack<EnteredDelimiter<'a>>)> {
@@ -238,6 +258,7 @@ impl<'a> Vertex<'a> {
             parents,
             lhs_parent_id: None,
             rhs_parent_id: None,
+            can_pop_either: false,
         }
     }
 }
@@ -270,11 +291,18 @@ pub enum Edge {
 impl Edge {
     pub fn cost(&self) -> u64 {
         match self {
+            // When we're at the end of a list, there's only one exit
+            // delimiter possibility, so the cost doesn't matter. We
+            // choose a non-zero number as it's easier to reason
+            // about.
             ExitDelimiterBoth => 1,
+            // Choose a higher value for exiting individually. This
+            // shouldn't matter since entering a novel delimiter is
+            // already more expensive than entering a matched
+            // delimiter, but be defensive.
             ExitDelimiterLHS | ExitDelimiterRHS => 2,
 
             // Matching nodes is always best.
-            // TODO: now we model parents correctly, do we need to track depth difference?
             UnchangedNode { depth_difference } => min(40, *depth_difference as u64 + 1),
             // Matching an outer delimiter is good.
             EnterUnchangedDelimiter { depth_difference } => 100 + min(40, *depth_difference as u64),
@@ -331,6 +359,7 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
                 Vertex {
                     lhs_syntax: lhs_parent.next_sibling(),
                     rhs_syntax: rhs_parent.next_sibling(),
+                    can_pop_either: can_pop_either_parent(&parents_next),
                     parents: parents_next,
                     lhs_parent_id: lhs_parent.parent().map(Syntax::id),
                     rhs_parent_id: rhs_parent.parent().map(Syntax::id),
@@ -350,6 +379,7 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
                 Vertex {
                     lhs_syntax: lhs_parent.next_sibling(),
                     rhs_syntax: v.rhs_syntax,
+                    can_pop_either: can_pop_either_parent(&parents_next),
                     parents: parents_next,
                     lhs_parent_id: lhs_parent.parent().map(Syntax::id),
                     rhs_parent_id: v.rhs_parent_id,
@@ -369,6 +399,7 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
                 Vertex {
                     lhs_syntax: v.lhs_syntax,
                     rhs_syntax: rhs_parent.next_sibling(),
+                    can_pop_either: can_pop_either_parent(&parents_next),
                     parents: parents_next,
                     lhs_parent_id: v.lhs_parent_id,
                     rhs_parent_id: rhs_parent.parent().map(Syntax::id),
@@ -393,6 +424,7 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
                     parents: v.parents.clone(),
                     lhs_parent_id: v.lhs_parent_id,
                     rhs_parent_id: v.rhs_parent_id,
+                    can_pop_either: v.can_pop_either,
                 },
             ));
             i += 1;
@@ -433,6 +465,7 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
                         parents: parents_next,
                         lhs_parent_id: Some(lhs_syntax.id()),
                         rhs_parent_id: Some(rhs_syntax.id()),
+                        can_pop_either: false,
                     },
                 ));
                 i += 1;
@@ -465,6 +498,7 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
                         parents: v.parents.clone(),
                         lhs_parent_id: v.lhs_parent_id,
                         rhs_parent_id: v.rhs_parent_id,
+                        can_pop_either: v.can_pop_either,
                     },
                 ));
                 i += 1;
@@ -488,6 +522,7 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
                         parents: v.parents.clone(),
                         lhs_parent_id: v.lhs_parent_id,
                         rhs_parent_id: v.rhs_parent_id,
+                        can_pop_either: v.can_pop_either,
                     },
                 ));
                 i += 1;
@@ -512,6 +547,7 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
                         parents: parents_next,
                         lhs_parent_id: Some(lhs_syntax.id()),
                         rhs_parent_id: v.rhs_parent_id,
+                        can_pop_either: true,
                     },
                 ));
                 i += 1;
@@ -527,6 +563,7 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
                             parents: v.parents.clone(),
                             lhs_parent_id: v.lhs_parent_id,
                             rhs_parent_id: v.rhs_parent_id,
+                            can_pop_either: v.can_pop_either,
                         },
                     ));
                     i += 1;
@@ -549,6 +586,7 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
                         parents: v.parents.clone(),
                         lhs_parent_id: v.lhs_parent_id,
                         rhs_parent_id: v.rhs_parent_id,
+                        can_pop_either: v.can_pop_either,
                     },
                 ));
                 i += 1;
@@ -573,6 +611,7 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
                         parents: parents_next,
                         lhs_parent_id: v.lhs_parent_id,
                         rhs_parent_id: Some(rhs_syntax.id()),
+                        can_pop_either: true,
                     },
                 ));
                 i += 1;
@@ -588,6 +627,7 @@ pub fn neighbours<'a>(v: &Vertex<'a>, buf: &mut [Option<(Edge, Vertex<'a>)>]) {
                             parents: v.parents.clone(),
                             lhs_parent_id: v.lhs_parent_id,
                             rhs_parent_id: v.rhs_parent_id,
+                            can_pop_either: v.can_pop_either,
                         },
                     ));
                     i += 1;
