@@ -35,6 +35,7 @@ pub enum Language {
     Jsx,
     OCaml,
     OCamlInterface,
+    Php,
     Python,
     Ruby,
     Rust,
@@ -57,7 +58,7 @@ pub fn guess(path: &Path, src: &str) -> Option<Language> {
     }
 
     match path.extension() {
-        Some(extension) => from_extension(extension),
+        Some(extension) => from_extension(extension, src),
         None => None,
     }
 }
@@ -162,7 +163,7 @@ fn from_name(path: &Path) -> Option<Language> {
     }
 }
 
-fn from_extension(extension: &OsStr) -> Option<Language> {
+fn from_extension(extension: &OsStr, src: &str) -> Option<Language> {
     match extension.to_string_lossy().borrow() {
         "sh" | "bash" | "bats" | "cgi" | "command" | "env" | "fcgi" | "ksh" | "sh.in" | "tmux"
         | "tool" | "zsh" => Some(Bash),
@@ -190,6 +191,7 @@ fn from_extension(extension: &OsStr) -> Option<Language> {
         "json" => Some(Json),
         "ml" => Some(OCaml),
         "mli" => Some(OCamlInterface),
+        "php" if !src.starts_with("<?hh") => Some(Php),
         "py" | "py3" | "pyi" | "bzl" => Some(Python),
         "rb" | "builder" | "spec" | "rake" => Some(Ruby),
         "rs" => Some(Rust),
