@@ -60,7 +60,7 @@ impl Hunk {
             deduped_lines.push((
                 if lhs_is_dupe { None } else { lhs_line },
                 if rhs_is_dupe { None } else { rhs_line },
-            ))
+            ));
         }
 
         Hunk {
@@ -513,120 +513,6 @@ pub fn matched_pos_to_hunks(lhs_mps: &[MatchedPos], rhs_mps: &[MatchedPos]) -> V
     lines_to_hunks(&matched_novel_lines(lhs_mps, rhs_mps), lhs_mps, rhs_mps)
 }
 
-/// Ensure that we don't miss any intermediate values.
-///
-/// Before:
-///
-/// ```text
-/// 1 11
-/// 3 14
-/// 4 --
-/// ```
-///
-/// After:
-///
-/// ```text
-/// 1 10
-/// 2 12 (choosing to align even though content doesn't match)
-/// - 13 (fix uneven gap)
-/// 3 14
-/// 4 -- (preserve outer gaps)
-/// ```
-pub fn ensure_contiguous(
-    lines: &[(Option<LineNumber>, Option<LineNumber>)],
-) -> Vec<(Option<LineNumber>, Option<LineNumber>)> {
-    let mut res: Vec<(Option<LineNumber>, Option<LineNumber>)> = vec![];
-
-    let mut lhs_min: Option<LineNumber> = None;
-    let mut rhs_min: Option<LineNumber> = None;
-
-    for (lhs_line, rhs_line) in lines {
-        if let Some(lhs_line) = lhs_line {
-            if let Some(lhs_prev_min) = lhs_min {
-                if lhs_prev_min.0 + 1 != lhs_line.0 {
-                    res.extend((lhs_prev_min.0 + 1..lhs_line.0).map(|l| (Some(l.into()), None)));
-                }
-            }
-            lhs_min = Some(*lhs_line);
-        }
-        if let Some(rhs_line) = rhs_line {
-            if let Some(rhs_prev_min) = rhs_min {
-                if rhs_prev_min.0 + 1 != rhs_line.0 {
-                    res.extend((rhs_prev_min.0 + 1..rhs_line.0).map(|r| (None, Some(r.into()))));
-                }
-            }
-            rhs_min = Some(*rhs_line);
-        }
-        res.push((*lhs_line, *rhs_line));
-    }
-
-    res
-}
-
-/// Before:
-///
-/// 10 --
-/// 11 --
-/// -- 20
-/// 12 21
-///
-/// After:
-///
-/// 10 20
-/// 11 --
-/// 12 21
-///
-/// The returned vec will contain no (None, None) pairs.
-pub fn compact_gaps(
-    items: &[(Option<LineNumber>, Option<LineNumber>)],
-) -> Vec<(Option<LineNumber>, Option<LineNumber>)> {
-    let mut res: Vec<(Option<LineNumber>, Option<LineNumber>)> = vec![];
-    // A vec of the most recent single-sided lines, e.g.
-    //
-    // 10 --
-    // 11 --
-    //
-    // All items must be Some on the same side.
-    let mut unpaired_lines: Vec<(Option<LineNumber>, Option<LineNumber>)> = vec![];
-
-    for (lhs_line, rhs_line) in items {
-        match (lhs_line, rhs_line) {
-            (Some(lhs_line), None) => {
-                match unpaired_lines.first() {
-                    Some((None, Some(rhs_line))) => {
-                        // We've found a line that can be compacted.
-                        res.push((Some(*lhs_line), Some(*rhs_line)));
-                        unpaired_lines.remove(0);
-                    }
-                    _ => {
-                        unpaired_lines.push((Some(*lhs_line), None));
-                    }
-                }
-            }
-            (None, Some(rhs_line)) => {
-                match unpaired_lines.first() {
-                    Some((Some(lhs_line), None)) => {
-                        // We've found a line that can be compacted.
-                        res.push((Some(*lhs_line), Some(*rhs_line)));
-                        unpaired_lines.remove(0);
-                    }
-                    _ => {
-                        unpaired_lines.push((None, Some(*rhs_line)));
-                    }
-                }
-            }
-            _ => {
-                res.extend(unpaired_lines);
-                unpaired_lines = vec![];
-                res.push((*lhs_line, *rhs_line));
-            }
-        }
-    }
-
-    res.extend(unpaired_lines);
-    res
-}
-
 fn either_side_equal(
     x: &(Option<LineNumber>, Option<LineNumber>),
     y: &(Option<LineNumber>, Option<LineNumber>),
@@ -651,22 +537,35 @@ pub fn matched_lines_for_hunk(
     matched_lines: &[(Option<LineNumber>, Option<LineNumber>)],
     hunk: &Hunk,
 ) -> Vec<(Option<LineNumber>, Option<LineNumber>)> {
-    // TODO: Use binary search instead.
-    let hunk_first = hunk.lines.first().expect("Hunks are non-empty");
-    let hunk_last = hunk.lines.last().expect("Hunks are non-empty");
+    let mut hunk_lhs_novel = hunk.novel_lhs.iter().copied().collect::<Vec<_>>();
+    hunk_lhs_novel.sort();
 
+    let mut hunk_rhs_novel = hunk.novel_rhs.iter().copied().collect::<Vec<_>>();
+    hunk_rhs_novel.sort();
+
+    let hunk_smallest = (
+        hunk_lhs_novel.first().copied(),
+        hunk_rhs_novel.first().copied(),
+    );
+    let hunk_largest = (
+        hunk_lhs_novel.last().copied(),
+        hunk_rhs_novel.last().copied(),
+    );
+
+    // TODO: Use binary search instead.
     let mut start_i = None;
     for (i, matched_line) in matched_lines.iter().enumerate() {
-        if either_side_equal(matched_line, hunk_first) {
+        if either_side_equal(matched_line, &hunk_smallest) {
             start_i = Some(i);
             break;
         }
     }
 
     let mut end_i = None;
-    for (i, matched_line) in matched_lines.iter().enumerate() {
-        if either_side_equal(matched_line, hunk_last) {
+    for (i, matched_line) in matched_lines.iter().enumerate().rev() {
+        if either_side_equal(matched_line, &hunk_largest) {
             end_i = Some(i + 1);
+            break;
         }
     }
 
@@ -791,39 +690,6 @@ mod tests {
         assert_eq!(hunks[0].lines, vec![(Some(0.into()), Some(0.into()))]);
     }
 
-    #[test]
-    fn test_compact_gaps() {
-        let res = compact_gaps(&[
-            (Some(0.into()), None),
-            (None, Some(0.into())),
-            (Some(1.into()), Some(1.into())),
-        ]);
-        assert_eq!(
-            res,
-            vec![
-                (Some(0.into()), Some(0.into())),
-                (Some(1.into()), Some(1.into())),
-            ]
-        )
-    }
-
-    #[test]
-    fn test_compact_gaps_with_larger_gap() {
-        let res = compact_gaps(&[
-            (Some(0.into()), None),
-            (Some(1.into()), None),
-            (None, Some(0.into())),
-            (None, Some(1.into())),
-        ]);
-        assert_eq!(
-            res,
-            vec![
-                (Some(0.into()), Some(0.into())),
-                (Some(1.into()), Some(1.into())),
-            ]
-        )
-    }
-
     #[test]
     fn test_matched_lines_for_hunk() {
         let matched_lines = &[
